package com.tzld.piaoquan.recommend.server.service.filter;

import com.alibaba.fastjson.JSONObject;
import com.google.common.base.Stopwatch;
import com.google.common.collect.Lists;
import com.tzld.piaoquan.recommend.server.common.ThreadPoolFactory;
import com.tzld.piaoquan.recommend.server.service.PreViewedService;
import com.tzld.piaoquan.recommend.server.service.ServiceBeanFactory;
import com.tzld.piaoquan.recommend.server.service.ViewedService;
import com.tzld.piaoquan.recommend.server.service.filter.strategy.*;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Slf4j
public abstract class AbstractFilterService {
    @Autowired
    private PreViewedService preViewedService;

    @Autowired
    private ViewedService viewedService;

    private ExecutorService pool = ThreadPoolFactory.filterPool();

    @Value("${newFilterGlobalSwitch:false}")
    private boolean newFilterGlobalSwitch;
    @Value("${newFilterAbExpCode:}")
    private String newFilterAbExpCode;


    private List<Long> viewFilterOld(FilterParam param) {
        // log.info("filterParam={}", JSONUtils.toJson(param));
        // 风险过滤
        List<Long> videoIds = filterWithRiskVideo(param.getRiskFilterFlag(),
                param.getAppType(), param.getRegionCode(), param.getAppRegionFiltered(), param.getVideosWithRisk(),
                param.getVideoIds(), param.getForceTruncation());
        // log.info("filterByRiskVideos videoIds={}", JSONUtils.toJson(videoIds));
        JSONObject obj = new JSONObject();
        obj.put("zhangbotest", "");
        obj.put("concurrent", param.concurrent);
        obj.put("notUsePreView", param.notUsePreView);
        obj.put("size1", videoIds.size());
        if (param.isNotUsePreView()) {
            ;
        } else {
            videoIds = filterByPreViewed(param.getAppType(), param.getMid(), videoIds);
        }
        obj.put("size2", videoIds.size());
        // log.info("filterByPreViewed videoIds={}", JSONUtils.toJson(videoIds));
        if (param.isConcurrent()) {
            videoIds = filterByViewedConcurrent(param.getAppType(), param.getMid(), param.getUid(), videoIds);
        } else {
            videoIds = filterByViewed(param.getAppType(), param.getMid(), param.getUid(), videoIds);
        }
        obj.put("size3", videoIds.size());
        log.info(obj.toString());
        // log.info("filterByViewed videoIds={}", JSONUtils.toJson(videoIds));
        return videoIds;
    }

    private List<Long> filterByViewedConcurrent(int appType, String mid, String uid, List<Long> videoIds) {
        // TODO uid为空时，还需要过滤么？
        if (StringUtils.isBlank(mid)
                || CollectionUtils.isEmpty(videoIds)) {
            return videoIds;
        }

//        int chunkSize = 20;
//        Collection<List<Long>> chunks = videoIds.stream()
//                .collect(Collectors.groupingBy(it -> it / chunkSize))
//                .values();

        int chunkSize = 20;
        List<List<Long>> chunks = new ArrayList<>();
        int size = videoIds.size();

        for (int i = 0; i < size; i += chunkSize) {
            int endIndex = Math.min(i + chunkSize, size);
            List<Long> chunk = videoIds.subList(i, endIndex);
            chunks.add(chunk);
        }


        CountDownLatch cdl = new CountDownLatch(chunks.size());
        List<Future<List<Long>>> futures = new ArrayList<>();
        for (final List<Long> ids : chunks) {
            Future<List<Long>> future = pool.submit(() ->
                    viewedService.filterViewedVideo(appType, mid, uid, ids));
            futures.add(future);
        }
        try {
            cdl.await(150, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("filter error", e);
            return null;
        }

        List<Long> result = new ArrayList<>();
        for (Future<List<Long>> f : futures) {
            try {
                result.addAll(f.get());
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }

        return result;


    }

    private List<Long> filterByViewed(int appType, String mid, String uid, List<Long> videoIds) {
        // TODO uid为空时，还需要过滤么？
        if (StringUtils.isBlank(mid)
                || CollectionUtils.isEmpty(videoIds)) {
            return videoIds;
        }
        return viewedService.filterViewedVideo(appType, mid, uid, videoIds);

    }

    private List<Long> filterByPreViewed(int appType, String mid, List<Long> videoIds) {

        if (StringUtils.isBlank(mid)) {
            return videoIds;
        }
        Set<Long> preViewedVideoIds = preViewedService.getVideoIds(appType, mid);
//        log.info("preViewedService.getVideoIds preViewedVideoIds={}, appType={}, mid={}",
//                JSONUtils.toJson(preViewedVideoIds), appType, mid);
        return videoIds.stream()
                .filter(l -> !preViewedVideoIds.contains(l))
                .collect(Collectors.toList());

    }

    private List<Long> filterWithRiskVideo(boolean riskFlag,
                                           int appType,
                                           String regionCode,
                                           Map<Integer, List<String>> rules,
                                           List<Long> videosWithRisk,
                                           List<Long> videoIds,
                                           int forceTruncation) {
        if (!riskFlag) {
            return this.truncation(videoIds, forceTruncation);
        }
        // 1 判断是否过滤,不展示的app+区域列表。
        boolean filterFlag;
        if (rules.containsKey(appType)) {
            filterFlag = false;
            if (rules.get(appType).contains(regionCode)) {
                filterFlag = true;
            }
        } else {
            filterFlag = true;
        }
        if (!filterFlag) {
            return this.truncation(videoIds, forceTruncation);
        }
        // 2 开始过滤。
        List<Long> videoIdNew = new ArrayList<>();
        for (Long videoId : videoIds) {
            if (!videosWithRisk.contains(videoId)) {
                videoIdNew.add(videoId);
            }
        }
        return this.truncation(videoIdNew, forceTruncation);
    }

    private List<Long> truncation(List<Long> videoIds, int forceTruncation) {
        if (forceTruncation == 0) {
            return videoIds;
        } else {
            return videoIds.subList(0, Math.min(forceTruncation, videoIds.size()));
        }
    }

    protected List<Long> viewFilter(FilterParam param) {
        boolean hit = newFilterGlobalSwitch
                || CommonCollectionUtils.contains(param.getAbExpCodes(), newFilterAbExpCode);
        if (hit) {
            return viewFilterNew(param);
        } else {
            return viewFilterOld(param);
        }
    }

    private List<Long> viewFilterNew(FilterParam param) {

        // hardcode : 风险过滤会做截断，所以先同步调用
        List<Long> riskVideoIds = filterWithRiskVideo(param.getRiskFilterFlag(),
                param.getAppType(), param.getRegionCode(), param.getAppRegionFiltered(), param.getVideosWithRisk(),
                param.getVideoIds(), param.getForceTruncation());
        param.setVideoIds(riskVideoIds);

        List<FilterStrategy> strategies = getStrategies(param);
        log.info("filter strategies {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies,
                o -> o.getClass().getSimpleName())));
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<List<Long>>> futures = new ArrayList<>();
        for (final FilterStrategy strategy : strategies) {
            Future<List<Long>> future = pool.submit(() -> {
                Stopwatch stopwatch = Stopwatch.createStarted();
                List<Long> result = strategy.filter(param);
                cdl.countDown();
                log.info("{} param {} result {} cost {} ms", strategy.getClass().getSimpleName(),
                        JSONUtils.toJson(param.getVideoIds()),
                        JSONUtils.toJson(result),
                        stopwatch.elapsed().toMillis());
                return result;
            });
            futures.add(future);
        }
        try {
            cdl.await(1000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("filter error", e);
            return null;
        }

        List<List<Long>> videoIds = new ArrayList<>();
        for (Future<List<Long>> f : futures) {
            try {
                videoIds.add(f.get());
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }
        if (CollectionUtils.isEmpty(videoIds)) {
            return Collections.emptyList();
        }
        List<Long> result = Lists.newArrayList(param.getVideoIds());
        for (int i = 0; i < videoIds.size(); ++i) {
            result.retainAll(videoIds.get(i));
        }
        log.info("filter result {}", JSONUtils.toJson(result));
        return result;
    }

    private List<FilterStrategy> getStrategies(FilterParam param) {
        List<FilterStrategy> strategies = new ArrayList<>();
        strategies.add(ServiceBeanFactory.getBean(PreViewedStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(ViewedStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(RecommendStatusStrategy.class));
        switch (param.getAppType()) {
            case 0:
            case 4:
                strategies.add(ServiceBeanFactory.getBean(AllowListStrategy.class));
                break;
            case 13:
                break;
            default:
                strategies.add(ServiceBeanFactory.getBean(AllowListStrategy.class));
                strategies.add(ServiceBeanFactory.getBean(TagStrategy.class));
                break;
        }

        return strategies;
    }

}
