package com.tzld.piaoquan.recommend.server.service.recall.strategy;

import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.filter.FilterParam;
import com.tzld.piaoquan.recommend.server.service.filter.FilterResult;
import com.tzld.piaoquan.recommend.server.service.filter.RegionFilterService;
import com.tzld.piaoquan.recommend.server.service.recall.FilterParamFactory;
import com.tzld.piaoquan.recommend.server.service.recall.RecallParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallStrategy;
import com.tzld.piaoquan.recommend.server.service.score.ScorerUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import java.util.stream.*;
import java.util.*;

import com.tzld.piaoquan.recommend.server.service.score4recall.ScorerPipeline4Recall;
import org.springframework.stereotype.Component;

/**
 * @author zhangbo
 */
@Slf4j
@Component
public class RegionRealtimeRecallStrategyV1 implements RecallStrategy {
    @Autowired
    private RegionFilterService filterService;
    @Override
    public List<Video> recall(RecallParam param) {
        // 1 获取省份key 放入参数map中
        String provinceCn = param.getProvince();
        if (provinceCn == null){
            provinceCn = "中国";
        }else{
            provinceCn = provinceCn.replaceAll("省$", "");
        }
        Map<String, String> param4Model = new HashMap<>(1);
        param4Model.put("region_province", provinceCn);
        // 2 通过model拿到召回list
        ScorerPipeline4Recall pipeline = ScorerUtils.getScorerPipeline4Recall("feeds_recall_config_region_v1.conf");
        List<List<Pair<Long, Double>>> results = pipeline.recall(param4Model);
        List<Pair<Long, Double>> result = results.get(0);
        for (int i=1; i<results.size(); ++i){
            result.addAll(results.get(i));
        }
        Map<Long, Double> videoMap = new LinkedHashMap<>();
        for (Pair<Long, Double> v: result){
            videoMap.put(v.getLeft(), v.getRight());
        }


        long t1 = new Long(System.currentTimeMillis());
//        int chunkSize = 25;
//        List<List<Long>> groupedKeys = groupKeys(videoMap, chunkSize);
//        List<Long> videoids = new ArrayList<>();
//        for (List<Long> tmp : groupedKeys){
//            FilterParam filterParam = FilterParamFactory.create(param, tmp);
//            filterParam.setForceTruncation(10000);
//            filterParam.setConcurrent(true);
//            filterParam.setUsePreView(false);
//            FilterResult filterResult = filterService.filter(filterParam);
//            if (filterResult != null && CollectionUtils.isNotEmpty(filterResult.getVideoIds())){
//                videoids.addAll(filterResult.getVideoIds());
//            }
//        }
        FilterParam filterParam = FilterParamFactory.create(param, Lists.newArrayList(videoMap.keySet()));
        filterParam.setForceTruncation(10000);
        filterParam.setConcurrent(true);
        filterParam.setNotUsePreView(false);
        FilterResult filterResult = filterService.filter(filterParam);
        long t2 = new Long(System.currentTimeMillis());
        JSONObject obj = new JSONObject();
        obj.put("name", "RegionRealtimeRecallStrategyV1");
        obj.put("filter_time", t2-t1);
        obj.put("provinceCn", provinceCn);
        obj.put("sizeOld", videoMap.size());
        List<Video> videosResult = new ArrayList<>();
        if (filterResult != null && CollectionUtils.isNotEmpty(filterResult.getVideoIds())) {
            obj.put("sizeNew", filterResult.getVideoIds().size());
            filterResult.getVideoIds().stream().forEach(vid -> {
                Video video = new Video();
                video.setVideoId(vid);
                video.setAbCode(param.getAbCode());
                video.setRovScore(videoMap.get(vid));
                video.setPushFrom(pushFrom());
                videosResult.add(video);
            });
        }
        log.info(obj.toString());
        Collections.sort(videosResult, Comparator.comparingDouble(o -> -o.getRovScore()));
        return videosResult;
    }
    public static final String PUSH_FORM = "recall_strategy_region_1h";
    @Override
    public String pushFrom(){
        return PUSH_FORM;
    }


    public static List<List<Long>> groupKeys(Map<Long, Double> videoMap, int groupSize) {
        List<List<Long>> result = new ArrayList<>();
        List<Long> keys = new ArrayList<>(videoMap.keySet());

        int size = keys.size();
        for (int i = 0; i < size; i += groupSize) {
            int endIndex = Math.min(i + groupSize, size);
            List<Long> group = keys.subList(i, endIndex);
            result.add(group);
        }

        return result;
    }

}
